"""
This module provide type-safe data window configuration
"""

from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Optional, Set, MutableMapping

import pprint
import toml

from redvox.common import date_time_utils as dtu


# defaults for configuration
DEFAULT_GAP_TIME_S: float = 0.25
DEFAULT_START_PADDING_S: float = 120.
DEFAULT_END_PADDING_S: float = 120.


@dataclass_json()
@dataclass
class DataWindowConfig:
    """
    Properties:
        input_directory: string, directory that contains the files to read data from.  REQUIRED
        station_ids: optional set of strings, list of station ids to filter on.
                        If empty or None, get any ids found in the input directory.  Default None
        start_datetime: optional datetime, start datetime of the window.
                        If None, uses the first timestamp of the filtered data.  Default None
        end_datetime: optional datetime, end datetime of the window.
                        If None, uses the last timestamp of the filtered data.  Default None
        start_padding_s: float, the amount of seconds to include before the start_datetime
                            when filtering data.  Default DEFAULT_START_PADDING_S
        end_padding_s: float, the amount of seconds to include after the end_datetime
                        when filtering data.  Default DEFAULT_END_PADDING_S
        gap_time_s: float, the minimum amount of seconds between data points that would indicate a gap.
                    Default DEFAULT_GAP_TIME_S
        apply_correction: bool, if True, update the timestamps in the data based on best station offset.  Default True
        structured_layout: bool, if True, the input_directory contains specially named and organized
                            directories of data.  Default True
    """
    input_directory: str
    station_ids: Optional[Set[str]] = None
    start_datetime: Optional[dtu.datetime] = None
    end_datetime: Optional[dtu.datetime] = None
    start_padding_s: float = DEFAULT_START_PADDING_S
    end_padding_s: float = DEFAULT_END_PADDING_S
    gap_time_s: float = DEFAULT_GAP_TIME_S
    apply_correction: bool = True
    structured_layout: bool = True

    @staticmethod
    def from_path(config_path: str) -> 'DataWindowConfig':
        try:
            with open(config_path, "r") as config_in:
                config_dict: MutableMapping = toml.load(config_in)
                # noinspection Mypy
                return DataWindowConfig.from_dict(config_dict)
        except Exception as e:
            print(f"Error loading configuration at: {config_path}")
            raise e

    def pretty(self) -> str:
        # noinspection Mypy
        return pprint.pformat(self.to_dict())



"""
This module creates specific time-bounded segments of data for users
"""
from typing import Optional, Set, List

import pandas as pd
import numpy as np

from redvox.common import date_time_utils as dtu
from redvox.common.station import Station
from redvox.common.station_reader_utils import ReadResult
from redvox.common.station_reader_utils import read_all_in_dir
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.api1000.wrapped_redvox_packet.sensors.image import ImageCodec


DEFAULT_GAP_TIME_S: float = 0.25        # default length of a gap in seconds
DEFAULT_START_PADDING_S: float = 120.   # default padding to start time of data in seconds
DEFAULT_END_PADDING_S: float = 120.     # default padding to end time of data in seconds
# default maximum number of points required to brute force calculate gap timestamps
DEFAULT_MAX_BRUTE_FORCE_GAP_TIMESTAMPS: int = 5000


class DataWindow:
    """
    Holds the data for a given time window; adds interpolated timestamps to fill gaps and pad start and end values
    Properties:
        input_directory: string, directory that contains the files to read data from.  REQUIRED
        station_ids: optional set of strings, list of station ids to filter on.
                        If empty or None, get any ids found in the input directory.  Default None
        start_datetime: optional datetime, start datetime of the window.
                        If None, uses the first timestamp of the filtered data.  Default None
        end_datetime: optional datetime, end datetime of the window.
                        If None, uses the last timestamp of the filtered data.  Default None
        start_padding_s: float, the amount of seconds to include before the start_datetime
                            when filtering data.  Default DEFAULT_START_PADDING_S
        end_padding_s: float, the amount of seconds to include after the end_datetime
                        when filtering data.  Default DEFAULT_END_PADDING_S
        gap_time_s: float, the minimum amount of seconds between data points that would indicate a gap.
                    Default DEFAULT_GAP_TIME_S
        apply_correction: bool, if True, update the timestamps in the data based on best station offset.  Default True
        structured_layout: bool, if True, the input_directory contains specially named and organized
                            directories of data.  Default True
        stations: optional ReadResult, the results of reading the data from input_directory
        debug: bool, if True, outputs additional information during initialization. Default False
    """
    def __init__(self, input_dir: str, station_ids: Optional[Set[str]] = None,
                 start_datetime: Optional[dtu.datetime] = None, end_datetime: Optional[dtu.datetime] = None,
                 start_padding_s: float = DEFAULT_START_PADDING_S, end_padding_s: float = DEFAULT_END_PADDING_S,
                 gap_time_s: float = DEFAULT_GAP_TIME_S, apply_correction: bool = True,
                 structured_layout: bool = True, debug: bool = False):
        """
        initialize the data window with params
        :param input_dir: string, directory that contains the files to read data from
        :param station_ids: optional set of strings, list of station ids to filter on.
                            If empty or None, get any ids found in the input directory.  Default None
        :param start_datetime: optional datetime, start datetime of the window.
                                If None, uses the first timestamp of the filtered data.  Default None
        :param end_datetime: optional datetime, end datetime of the window.
                                If None, uses the last timestamp of the filtered data.  Default None
        :param start_padding_s: float, the amount of seconds to include before the start_datetime
                                when filtering data.  Default DEFAULT_START_PADDING_S
        :param end_padding_s: float, the amount of seconds to include after the end_datetime
                                when filtering data.  Default DEFAULT_END_PADDING_S
        :param gap_time_s: float, the minimum amount of seconds between data points that would indicate a gap.
                            Default DEFAULT_GAP_TIME_S
        :param apply_correction: bool, if True, update the timestamps in the data based on best station offset.
                                    Default True
        :param structured_layout: bool, if True, the input_directory contains specially named and organized
                                    directories of data.  Default True
        :param debug: bool, if True, outputs warnings and additional information, default False
        """
        self.input_directory: str = input_dir
        self.station_ids: Optional[Set[str]] = station_ids
        self.start_datetime: Optional[dtu.datetime] = start_datetime
        self.end_datetime: Optional[dtu.datetime] = end_datetime
        self.start_padding_s: float = start_padding_s
        self.end_padding_s: float = end_padding_s
        self.gap_time_s: float = gap_time_s
        self.apply_correction: bool = apply_correction
        self.structured_layout: bool = structured_layout
        self.debug: bool = debug
        start_time = self._pad_start_datetime_s()
        if np.isnan(start_time):
            start_time = None
        else:
            start_time = int(start_time)
        end_time = self._pad_end_datetime_s()
        if np.isnan(end_time):
            end_time = None
        else:
            end_time = int(end_time)
        self.stations: ReadResult = read_all_in_dir(self.input_directory, start_time, end_time,
                                                    self.station_ids, self.structured_layout)
        if self.station_ids is None or len(self.station_ids) == 0:
            self.station_ids = self.stations.get_all_station_ids()
        else:
            self.check_valid_ids()
        self.create_data_window()

    def _has_time_window(self) -> bool:
        """
        Returns true if there is a start or end datetime in the settings
        :return: True if start_datetime or end_datetime exists
        """
        return self.start_datetime is not None or self.end_datetime is not None

    def _pad_start_datetime_s(self) -> float:
        """
        apply padding to the start datetime
        :return: padded start datetime as seconds since epoch UTC or np.nan if start datetime is undefined
        """
        if self.start_datetime is None:
            return np.nan
        return dtu.datetime_to_epoch_seconds_utc(self.start_datetime) - self.start_padding_s

    def _pad_end_datetime_s(self) -> float:
        """
        apply padding to the end datetime
        :return: padded end datetime as seconds since epoch UTC or np.nan if end datetime is undefined
        """
        if self.end_datetime is None:
            return np.nan
        return dtu.datetime_to_epoch_seconds_utc(self.end_datetime) + self.end_padding_s

    def get_station(self, station: str) -> Optional[Station]:
        """
        :param station: the station id to search for
        :return: A single station based on the id given or None if the station doesn't exist
        """
        return self.stations.get_station(station)

    def get_all_stations(self) -> List[Station]:
        """
        :return: A list of all stations in the object
        """
        return self.stations.get_all_stations()

    def correct_timestamps(self):
        """
        update the timestamps in all stations
        """
        if self.apply_correction:
            for station in self.stations.station_id_uuid_to_stations.values():
                station.update_timestamps()

    def check_valid_ids(self):
        """
        searches the data window station_ids for any ids not in the data collected
        """
        for ids in self.station_ids:
            if not self.stations.check_for_id(ids) and self.debug:
                print(f"WARNING: Requested {ids} but there is no data to read for that station")

    def create_window_in_sensors(self, station: Station, start_date_timestamp: float, end_date_timestamp: float):
        """
        truncate the sensors in the station to only contain data from start_date_timestamp to end_date_timestamp
        returns nothing, updates the station in place
        :param station: station object to truncate sensors of
        :param start_date_timestamp: float, timestamp in microseconds since epoch UTC of start of window
        :param end_date_timestamp: float, timestamp in microseconds since epoch UTC of end of window
        """
        gap_time_micros = dtu.seconds_to_microseconds(self.gap_time_s)
        station_id = station.station_metadata.station_id
        for sensor_type, sensor in station.station_data.items():
            # calculate the sensor's sample interval, std sample interval and sample rate of all data
            sensor.organize_and_update_stats()
            # get only the timestamps between the start and end timestamps
            df_timestamps = sensor.data_timestamps()
            if len(df_timestamps) > 0:
                window_indices = np.where((start_date_timestamp <= df_timestamps) &
                                          (df_timestamps <= end_date_timestamp))[0]
                # check if all the samples have been cut off
                if len(window_indices) < 1:
                    if self.debug:
                        print(f"WARNING: Data window for {station_id} {sensor_type.name} "
                              f"sensor has truncated all data points")
                else:
                    sensor.data_df = sensor.data_df.iloc[window_indices].reset_index(drop=True)
                    if sensor.is_sample_interval_invalid():
                        if self.debug:
                            print(f"WARNING: Cannot fill gaps or pad {station_id} {sensor_type.name} "
                                  f"sensor; it has undefined sample interval and sample rate!")
                    else:  # GAP FILL and PAD DATA
                        sample_interval_micros = dtu.seconds_to_microseconds(sensor.sample_interval_s)
                        sensor.data_df = fill_gaps(sensor.data_df, sample_interval_micros +
                                                   dtu.seconds_to_microseconds(sensor.sample_interval_std_s),
                                                   gap_time_micros, DEFAULT_MAX_BRUTE_FORCE_GAP_TIMESTAMPS)
                        sensor.data_df = pad_data(start_date_timestamp, end_date_timestamp, sensor.data_df,
                                                  sample_interval_micros)
            elif self.debug:
                print(f"WARNING: Data window for {station_id} {sensor_type.name} sensor has no data points!")

    def truncate_metadata(self, station: Station, start_date_timestamp: float, end_date_timestamp: float):
        """
        truncates and updates the metadata of the station to only be from start_date_timestamp to end_date_timestamp.
        Returns nothing; updates metadata in place
        :param station: station object to update
        :param start_date_timestamp: float, timestamp in microseconds since epoch UTC of start of window
        :param end_date_timestamp: float, timestamp in microseconds since epoch UTC of end of window
        """
        # station.packet_gap_detector(self.gap_time_s)
        station.packet_data = [p for p in station.packet_data
                               if p.data_end_timestamp > start_date_timestamp and
                               p.data_start_timestamp < end_date_timestamp]
        if station.has_location_data():
            # anything with 0 altitude is likely a network provided location
            station.location_sensor().data_df.loc[(station.location_sensor().data_df["altitude"] == 0),
                                                  "location_provider"] = LocationProvider.NETWORK
        station.update_station_location_metadata(start_date_timestamp, end_date_timestamp)
        station.station_metadata.timing_data.episode_start_timestamp_s = \
            dtu.microseconds_to_seconds(start_date_timestamp)
        station.station_metadata.timing_data.episode_end_timestamp_s = dtu.microseconds_to_seconds(end_date_timestamp)
        station.station_metadata.timing_data.station_first_data_timestamp = \
            station.audio_sensor().first_data_timestamp()

    def create_data_window(self):
        """
        updates the data window to contain only the data within the window parameters
        stations without audio or any data within the window are removed
        """
        ids_to_pop = []
        for station in self.stations.get_all_stations():
            ids_to_pop = check_audio_data(station, ids_to_pop, self.debug)
            # apply time correction
            if self.apply_correction:
                station.update_timestamps()
            # set the window start and end if they were specified, otherwise use the bounds of the data
            if self.start_datetime:
                start_datetime = dtu.datetime_to_epoch_microseconds_utc(self.start_datetime)
            else:
                start_datetime = station.audio_sensor().first_data_timestamp()
            if self.end_datetime:
                end_datetime = dtu.datetime_to_epoch_microseconds_utc(self.end_datetime)
            else:
                end_datetime = station.audio_sensor().last_data_timestamp()
            # TRUNCATE!
            self.create_window_in_sensors(station, start_datetime, end_datetime)
            self.truncate_metadata(station, start_datetime, end_datetime)
            ids_to_pop = check_audio_data(station, ids_to_pop, self.debug)
        # remove any stations that don't have audio data
        for ids in ids_to_pop:
            self.stations.pop_station(ids)


def check_audio_data(station: Station, ids_to_remove: List[str], debug: bool = False) -> List[str]:
    """
    check if the station has audio data; if it does not, update the list of stations to remove
    :param station: station object to check for audio data
    :param ids_to_remove: list of strings, the station ids to remove from the data window
    :param debug: bool, if True, output warning message, default False
    :return: an updated list of station ids to remove from the data window
    """
    station_id = station.station_metadata.station_id
    if not station.has_audio_data():
        if debug:
            print(f"WARNING: {station_id} doesn't have any audio data to read")
        ids_to_remove.append(station_id)
    return ids_to_remove


def pad_data(expected_start: float, expected_end: float, data_df: pd.DataFrame,
             sample_interval_micros: float) -> pd.DataFrame:
    """
    Pad the start and end of the dataframe with np.nan
    :param expected_start: timestamp indicating start time of the data to pad from
    :param expected_end: timestamp indicating end time of the data to pad from
    :param data_df: dataframe with timestamps as column "timestamps"
    :param sample_interval_micros: constant sample interval in microseconds
    :return: dataframe padded with np.nans in front and back to meet full size of expected start and end
    """
    # extract the necessary information to pad the data
    data_time_stamps = data_df["timestamps"].to_numpy()
    first_data_timestamp = data_time_stamps[0]
    last_data_timestamp = data_time_stamps[-1]
    result_df = data_df.copy()
    # FRONT/END GAP FILL!  calculate the audio samples missing based on inputs
    if expected_start < first_data_timestamp:
        start_diff = first_data_timestamp - expected_start
        num_missing_samples = np.ceil(start_diff / sample_interval_micros)
        # add the gap data to the result dataframe
        result_df = result_df.append(create_dataless_timestamps_df(expected_start - sample_interval_micros,
                                                                   sample_interval_micros, data_df.columns,
                                                                   num_missing_samples), ignore_index=True)
    if expected_end > last_data_timestamp:
        last_diff = expected_end - last_data_timestamp
        num_missing_samples = np.ceil(last_diff / sample_interval_micros)
        # add the gap data to the result dataframe
        result_df = result_df.append(create_dataless_timestamps_df(expected_end + sample_interval_micros,
                                                                   sample_interval_micros, data_df.columns,
                                                                   num_missing_samples, True), ignore_index=True)
    return result_df.sort_values("timestamps", ignore_index=True)


def fill_gaps(data_df: pd.DataFrame, sample_interval_micros: float, gap_time_micros: float,
              num_points_to_brute_force: int = DEFAULT_MAX_BRUTE_FORCE_GAP_TIMESTAMPS) -> pd.DataFrame:
    """
    fills gaps in the dataframe with np.nan by interpolating timestamps based on the mean expected sample interval
    :param data_df: dataframe with timestamps as column "timestamps"
    :param sample_interval_micros: float, sample interval in microseconds
    :param gap_time_micros: float, minimum amount of microseconds between data points that would indicate a gap
    :param num_points_to_brute_force: int, maximum number of points to calculate when filling a gap
    :return: dataframe without gaps
    """
    # extract the necessary information to compute gap size and gap timestamps
    data_time_stamps = data_df["timestamps"].to_numpy()
    first_data_timestamp = data_time_stamps[0]
    last_data_timestamp = data_time_stamps[-1]
    data_duration_micros = last_data_timestamp - first_data_timestamp
    num_points = len(data_time_stamps)
    # add one to calculation to include the last timestamp
    expected_num_points = np.ceil(data_duration_micros / sample_interval_micros) + 1
    # gap duration cannot be less than sample interval
    gap_time_micros = np.max([sample_interval_micros, gap_time_micros])
    result_df = data_df.copy()
    # if there are less points than our expected amount, we have gaps to fill
    if num_points < expected_num_points:
        # if the data we're looking at is short enough, we can start comparing points
        if num_points < num_points_to_brute_force:
            # look at every timestamp difference
            timestamp_diffs = np.diff(data_time_stamps)
            for index in np.where(timestamp_diffs > gap_time_micros)[0]:
                # calc samples to add, subtracting 1 to prevent copying last timestamp
                num_new_samples = np.ceil(timestamp_diffs[index] / sample_interval_micros) - 1
                if timestamp_diffs[index] > gap_time_micros and num_new_samples > 0:
                    # add the gap data to the result dataframe
                    result_df = result_df.append(
                        create_dataless_timestamps_df(data_time_stamps[index], sample_interval_micros,
                                                      data_df.columns, num_new_samples), ignore_index=True)
                    if len(result_df) >= expected_num_points:
                        break  # stop the for loop execution when enough points are added
        else:
            # too many points to check, divide and conquer using recursion!
            half_samples = int(num_points / 2)
            first_data_df = data_df.iloc[:half_samples].copy().reset_index(drop=True)
            second_data_df = data_df.iloc[half_samples:].copy().reset_index(drop=True)
            # give half the samples to each recursive call
            first_data_df = fill_gaps(first_data_df, sample_interval_micros, gap_time_micros, num_points_to_brute_force)
            second_data_df = fill_gaps(second_data_df, sample_interval_micros, gap_time_micros,
                                       num_points_to_brute_force)
            result_df = first_data_df.append(second_data_df, ignore_index=True)
    return result_df.sort_values("timestamps", ignore_index=True)


def create_dataless_timestamps_df(start_timestamp: float, sample_interval_micros: float, columns: pd.Index,
                                  num_samples_to_add: int, add_to_start: bool = False) -> pd.DataFrame:
    """
    Creates an empty dataframe with num_samples_to_add timestamps, using columns as the columns
    the first timestamp created is 1 sample_interval_s from the start_timestamp
    :param start_timestamp: float, timestamp in microseconds since epoch UTC to start calculating other timestamps from
    :param sample_interval_micros: float, fixed sample interval in microseconds since epoch UTC
    :param columns: dataframe index, the non-timestamp columns of the dataframe
    :param num_samples_to_add: int, the number of timestamps to create
    :param add_to_start: bool, if True, subtracts sample_interval_s from start_timestamp, default False
    :return: dataframe with timestamps and no data
    """
    if add_to_start:
        sample_interval_micros = -sample_interval_micros
    new_timestamps = start_timestamp + np.arange(1, num_samples_to_add + 1) * sample_interval_micros
    empty_df = pd.DataFrame([], columns=columns)
    for column_index in columns:
        if column_index == "timestamps":
            empty_df[column_index] = new_timestamps
        elif column_index == "location_provider":
            empty_df[column_index] = LocationProvider.UNKNOWN
        elif column_index == "image_codec":
            empty_df[column_index] = ImageCodec.UNKNOWN
        else:
            empty_df[column_index] = np.nan
    return empty_df



"""
Defines generic station objects for API-independent analysis
all timestamps are integers in microseconds unless otherwise stated
"""
from typing import List, Dict, Optional

import redvox.common.date_time_utils as dtu
from redvox.common.sensor_data import SensorData, SensorType
from redvox.common.station_utils import StationKey, StationMetadata, DataPacket, StationLocation, \
    station_location_from_data


class Station:
    """
    generic station for api-independent stuff
    Properties:
        station_metadata: StationMetadata
        station_data: dict, all the data associated with this station, default empty dict
        packet_data: list, all DataPacket metadata associated with this station, default empty list
        station_key: Tuple of str, str, float, a unique combination of three values defining the station
    """

    def __init__(self, metadata: StationMetadata, data: Optional[Dict[SensorType, SensorData]] = None,
                 packets: Optional[List[DataPacket]] = None):
        """
        initialize Station
        :param metadata: the station's metadata
        :param data: the station's sensors' data, default None (value is converted to empty dict)
        :param packets: the packets that the data came from, default None (value is converted to empty list)
        """
        self.station_metadata: StationMetadata = metadata
        if data:
            self.station_data: Dict[SensorType, SensorData] = data
        else:
            self.station_data: Dict[SensorType, SensorData] = {}
        # todo add event streams as dict[str, list[event]] i.e movement: [accelerometer, gyroscope, etc]
        if packets:
            self.packet_data: List[DataPacket] = packets
        else:
            self.packet_data: List[DataPacket] = []
        # todo: assert station key is valid
        self.station_key = StationKey(self.station_metadata.station_id, self.station_metadata.station_uuid,
                                      self.station_metadata.timing_data.station_start_timestamp)

    def append_station(self, new_station: 'Station'):
        """
        append a new station to the current station
        :param new_station: Station to append to current station
        """
        if new_station.station_metadata.station_id == self.station_metadata.station_id:
            self.append_station_data(new_station.station_data)
            self.packet_data.extend(new_station.packet_data)
        # todo: regenerate the metadata when adding new station data

    def append_station_data(self, new_station_data: Dict[SensorType, SensorData]):
        """
        append new station data to existing station data
        :param new_station_data: the dictionary of data to add
        """
        for sensor_type, sensor_data in new_station_data.items():
            self.append_sensor(sensor_type, sensor_data)

    def append_sensor(self, sensor_type: SensorType, sensor_data: SensorData):
        """
        append sensor data to an existing sensor_type or add a new sensor to the dictionary
        :param sensor_type: the sensor to append to
        :param sensor_data: the data to append
        """
        if sensor_type in self.station_data.keys():
            self.station_data[sensor_type] = self.station_data[sensor_type].append_data(sensor_data.data_df)
        else:
            self._add_sensor(sensor_type, sensor_data)

    def _delete_sensor(self, sensor_type: SensorType):
        """
        removes a sensor from the data packet if it exists
        :param sensor_type: the sensor to remove
        """
        if sensor_type in self.station_data.keys():
            self.station_data.pop(sensor_type)

    def _add_sensor(self, sensor_type: SensorType, sensor: SensorData):
        """
        adds a sensor to the sensor_data_dict
        :param sensor_type: the type of sensor to add
        :param sensor: the sensor data to add
        """
        if sensor_type in self.station_data.keys():
            raise ValueError(f"Cannot add sensor type ({sensor_type.name}) that already exists in packet!")
        else:
            self.station_data[sensor_type] = sensor

    def has_audio_sensor(self) -> bool:
        """
        check if audio sensor is in sensor_data_dict
        :return: True if audio sensor exists
        """
        return SensorType.AUDIO in self.station_data.keys()

    def has_audio_data(self) -> bool:
        """
        check if the audio sensor has any data
        :return: True if audio sensor has any data
        """
        return self.has_audio_sensor() and self.audio_sensor().num_samples() > 0

    def audio_sensor(self) -> Optional[SensorData]:
        """
        return the audio sensor if it exists
        :return: audio sensor if it exists, None otherwise
        """
        if self.has_audio_sensor():
            return self.station_data[SensorType.AUDIO]
        return None

    def set_audio_sensor(self, audio_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the audio sensor; can remove audio sensor by passing None
        :param audio_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_audio_sensor():
            self._delete_sensor(SensorType.AUDIO)
        if audio_sensor is not None:
            self._add_sensor(SensorType.AUDIO, audio_sensor)
        return self

    def has_location_sensor(self) -> bool:
        """
        check if location sensor is in sensor_data_dict
        :return: True if location sensor exists
        """
        return SensorType.LOCATION in self.station_data.keys()

    def has_location_data(self) -> bool:
        """
        check if the location sensor has any data
        :return: True if location sensor has any data
        """
        return self.has_location_sensor() and self.location_sensor().num_samples() > 0

    def location_sensor(self) -> Optional[SensorData]:
        """
        return the location sensor if it exists
        :return: location sensor if it exists, None otherwise
        """
        if self.has_location_sensor():
            return self.station_data[SensorType.LOCATION]
        return None

    def set_location_sensor(self, loc_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the location sensor; can remove location sensor by passing None
        :param loc_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_location_sensor():
            self._delete_sensor(SensorType.LOCATION)
        if loc_sensor is not None:
            self._add_sensor(SensorType.LOCATION, loc_sensor)
        return self

    def has_accelerometer_sensor(self) -> bool:
        """
        check if accelerometer sensor is in sensor_data_dict
        :return: True if accelerometer sensor exists
        """
        return SensorType.ACCELEROMETER in self.station_data.keys()

    def has_accelerometer_data(self) -> bool:
        """
        check if the accelerometer sensor has any data
        :return: True if accelerometer sensor has any data
        """
        return self.has_accelerometer_sensor() and self.accelerometer_sensor().num_samples() > 0

    def accelerometer_sensor(self) -> Optional[SensorData]:
        """
        return the accelerometer sensor if it exists
        :return: accelerometer sensor if it exists, None otherwise
        """
        if self.has_accelerometer_sensor():
            return self.station_data[SensorType.ACCELEROMETER]
        return None

    def set_accelerometer_sensor(self, acc_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the accelerometer sensor; can remove accelerometer sensor by passing None
        :param acc_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_accelerometer_sensor():
            self._delete_sensor(SensorType.ACCELEROMETER)
        if acc_sensor is not None:
            self._add_sensor(SensorType.ACCELEROMETER, acc_sensor)
        return self

    def has_magnetometer_sensor(self) -> bool:
        """
        check if magnetometer sensor is in sensor_data_dict
        :return: True if magnetometer sensor exists
        """
        return SensorType.MAGNETOMETER in self.station_data.keys()

    def has_magnetometer_data(self) -> bool:
        """
        check if the magnetometer sensor has any data
        :return: True if magnetometer sensor has any data
        """
        return self.has_magnetometer_sensor() and self.magnetometer_sensor().num_samples() > 0

    def magnetometer_sensor(self) -> Optional[SensorData]:
        """
        return the magnetometer sensor if it exists
        :return: magnetometer sensor if it exists, None otherwise
        """
        if self.has_magnetometer_sensor():
            return self.station_data[SensorType.MAGNETOMETER]
        return None

    def set_magnetometer_sensor(self, mag_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the magnetometer sensor; can remove magnetometer sensor by passing None
        :param mag_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_magnetometer_sensor():
            self._delete_sensor(SensorType.MAGNETOMETER)
        if mag_sensor is not None:
            self._add_sensor(SensorType.MAGNETOMETER, mag_sensor)
        return self

    def has_gyroscope_sensor(self) -> bool:
        """
        check if gyroscope sensor is in sensor_data_dict
        :return: True if gyroscope sensor exists
        """
        return SensorType.GYROSCOPE in self.station_data.keys()

    def has_gyroscope_data(self) -> bool:
        """
        check if the gyroscope sensor has any data
        :return: True if gyroscope sensor has any data
        """
        return self.has_gyroscope_sensor() and self.gyroscope_sensor().num_samples() > 0

    def gyroscope_sensor(self) -> Optional[SensorData]:
        """
        return the gyroscope sensor if it exists
        :return: gyroscope sensor if it exists, None otherwise
        """
        if self.has_gyroscope_sensor():
            return self.station_data[SensorType.GYROSCOPE]
        return None

    def set_gyroscope_sensor(self, gyro_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the gyroscope sensor; can remove gyroscope sensor by passing None
        :param gyro_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_gyroscope_sensor():
            self._delete_sensor(SensorType.GYROSCOPE)
        if gyro_sensor is not None:
            self._add_sensor(SensorType.GYROSCOPE, gyro_sensor)
        return self

    def has_barometer_sensor(self) -> bool:
        """
        check if barometer sensor is in sensor_data_dict
        :return: True if barometer sensor exists
        """
        return SensorType.PRESSURE in self.station_data.keys()

    def has_barometer_data(self) -> bool:
        """
        check if the barometer sensor has any data
        :return: True if barometer sensor has any data
        """
        return self.has_barometer_sensor() and self.barometer_sensor().num_samples() > 0

    def barometer_sensor(self) -> Optional[SensorData]:
        """
        return the barometer sensor if it exists
        :return: barometer sensor if it exists, None otherwise
        """
        if self.has_barometer_sensor():
            return self.station_data[SensorType.PRESSURE]
        return None

    def set_barometer_sensor(self, bar_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the barometer sensor; can remove barometer sensor by passing None
        :param bar_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_barometer_sensor():
            self._delete_sensor(SensorType.PRESSURE)
        if bar_sensor is not None:
            self._add_sensor(SensorType.PRESSURE, bar_sensor)
        return self

    def has_light_sensor(self) -> bool:
        """
        check if light sensor is in sensor_data_dict
        :return: True if light sensor exists
        """
        return SensorType.LIGHT in self.station_data.keys()

    def has_light_data(self) -> bool:
        """
        check if the light sensor has any data
        :return: True if light sensor has any data
        """
        return self.has_light_sensor() and self.light_sensor().num_samples() > 0

    def light_sensor(self) -> Optional[SensorData]:
        """
        return the light sensor if it exists
        :return: light sensor if it exists, None otherwise
        """
        if self.has_light_sensor():
            return self.station_data[SensorType.LIGHT]
        return None

    def set_light_sensor(self, light_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the light sensor; can remove light sensor by passing None
        :param light_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_light_sensor():
            self._delete_sensor(SensorType.LIGHT)
        if light_sensor is not None:
            self._add_sensor(SensorType.LIGHT, light_sensor)
        return self

    def has_infrared_sensor(self) -> bool:
        """
        check if infrared sensor is in sensor_data_dict
        :return: True if infrared sensor exists
        """
        return SensorType.INFRARED in self.station_data.keys()

    def has_infrared_data(self) -> bool:
        """
        check if the infrared sensor has any data
        :return: True if infrared sensor has any data
        """
        return self.has_infrared_sensor() and self.infrared_sensor().num_samples() > 0

    def infrared_sensor(self) -> Optional[SensorData]:
        """
        return the infrared sensor if it exists
        :return: infrared sensor if it exists, None otherwise
        """
        if self.has_infrared_sensor():
            return self.station_data[SensorType.INFRARED]
        return None

    def set_infrared_sensor(self, infrd_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the infrared sensor; can remove infrared sensor by passing None
        :param infrd_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_infrared_sensor():
            self._delete_sensor(SensorType.INFRARED)
        if infrd_sensor is not None:
            self._add_sensor(SensorType.INFRARED, infrd_sensor)
        return self

    def has_image_sensor(self) -> bool:
        """
        check if image sensor is in sensor_data_dict
        :return: True if image sensor exists
        """
        return SensorType.IMAGE in self.station_data.keys()

    def has_image_data(self) -> bool:
        """
        check if the image sensor has any data
        :return: True if image sensor has any data
        """
        return self.has_image_sensor() and self.image_sensor().num_samples() > 0

    def image_sensor(self) -> Optional[SensorData]:
        """
        return the image sensor if it exists
        :return: image sensor if it exists, None otherwise
        """
        if self.has_image_sensor():
            return self.station_data[SensorType.IMAGE]
        return None

    def set_image_sensor(self, img_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the image sensor; can remove image sensor by passing None
        :param img_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_image_sensor():
            self._delete_sensor(SensorType.IMAGE)
        if img_sensor is not None:
            self._add_sensor(SensorType.IMAGE, img_sensor)
        return self

    def has_ambient_temperature_sensor(self) -> bool:
        """
        check if ambient temperature sensor is in sensor_data_dict
        :return: True if ambient temperature sensor exists
        """
        return SensorType.TEMPERATURE in self.station_data.keys()

    def has_ambient_temperature_data(self) -> bool:
        """
        check if the ambient temperature sensor has any data
        :return: True if ambient temperature sensor has any data
        """
        return self.has_ambient_temperature_sensor() and self.ambient_temperature_sensor().num_samples() > 0

    def ambient_temperature_sensor(self) -> Optional[SensorData]:
        """
        return the ambient temperature sensor if it exists
        :return: image ambient temperature if it exists, None otherwise
        """
        if self.has_ambient_temperature_sensor():
            return self.station_data[SensorType.TEMPERATURE]
        return None

    def set_ambient_temperature_sensor(self, amtemp_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the ambient temperature sensor; can remove ambient temperature sensor by passing None
        :param amtemp_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_ambient_temperature_sensor():
            self._delete_sensor(SensorType.TEMPERATURE)
        if amtemp_sensor is not None:
            self._add_sensor(SensorType.TEMPERATURE, amtemp_sensor)
        return self

    def has_gravity_sensor(self) -> bool:
        """
        check if gravity sensor is in sensor_data_dict
        :return: True if gravity sensor exists
        """
        return SensorType.GRAVITY in self.station_data.keys()

    def has_gravity_data(self) -> bool:
        """
        check if the gravity sensor has any data
        :return: True if gravity sensor has any data
        """
        return self.has_gravity_sensor() and self.gravity_sensor().num_samples() > 0

    def gravity_sensor(self) -> Optional[SensorData]:
        """
        return the gravity sensor if it exists
        :return: gravity sensor if it exists, None otherwise
        """
        if self.has_gravity_sensor():
            return self.station_data[SensorType.GRAVITY]
        return None

    def set_gravity_sensor(self, grav_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the gravity sensor; can remove gravity sensor by passing None
        :param grav_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_gravity_sensor():
            self._delete_sensor(SensorType.GRAVITY)
        if grav_sensor is not None:
            self._add_sensor(SensorType.GRAVITY, grav_sensor)
        return self

    def has_linear_acceleration_sensor(self) -> bool:
        """
        check if linear acceleration sensor is in sensor_data_dict
        :return: True if linear acceleration sensor exists
        """
        return SensorType.LINEAR_ACCELERATION in self.station_data.keys()

    def has_linear_acceleration_data(self) -> bool:
        """
        check if the linear acceleration sensor has any data
        :return: True if linear acceleration sensor has any data
        """
        return self.has_linear_acceleration_sensor() and self.linear_acceleration_sensor().num_samples() > 0

    def linear_acceleration_sensor(self) -> Optional[SensorData]:
        """
        return the linear acceleration sensor if it exists
        :return: linear acceleration sensor if it exists, None otherwise
        """
        if self.has_linear_acceleration_sensor():
            return self.station_data[SensorType.LINEAR_ACCELERATION]
        return None

    def set_linear_acceleration_sensor(self, linacc_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the linear acceleration sensor; can remove linear acceleration sensor by passing None
        :param linacc_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_linear_acceleration_sensor():
            self._delete_sensor(SensorType.LINEAR_ACCELERATION)
        if linacc_sensor is not None:
            self._add_sensor(SensorType.LINEAR_ACCELERATION, linacc_sensor)
        return self

    def has_orientation_sensor(self) -> bool:
        """
        check if orientation sensor is in sensor_data_dict
        :return: True if orientation sensor exists
        """
        return SensorType.ORIENTATION in self.station_data.keys()

    def has_orientation_data(self) -> bool:
        """
        check if the orientation sensor has any data
        :return: True if orientation sensor has any data
        """
        return self.has_orientation_sensor() and self.orientation_sensor().num_samples() > 0

    def orientation_sensor(self) -> Optional[SensorData]:
        """
        return the orientation sensor if it exists
        :return: orientation sensor if it exists, None otherwise
        """
        if self.has_orientation_sensor():
            return self.station_data[SensorType.ORIENTATION]
        return None

    def set_orientation_sensor(self, orient_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the orientation sensor; can remove orientation sensor by passing None
        :param orient_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_orientation_sensor():
            self._delete_sensor(SensorType.ORIENTATION)
        if orient_sensor is not None:
            self._add_sensor(SensorType.ORIENTATION, orient_sensor)
        return self

    def has_proximity_sensor(self) -> bool:
        """
        check if proximity sensor is in sensor_data_dict
        :return: True if proximity sensor exists
        """
        return SensorType.PROXIMITY in self.station_data.keys()

    def has_proximity_data(self) -> bool:
        """
        check if the proximity sensor has any data
        :return: True if proximity sensor has any data
        """
        return self.has_proximity_sensor() and self.proximity_sensor().num_samples() > 0

    def proximity_sensor(self) -> Optional[SensorData]:
        """
        return the proximity sensor if it exists
        :return: proximity sensor if it exists, None otherwise
        """
        if self.has_proximity_sensor():
            return self.station_data[SensorType.PROXIMITY]
        return None

    def set_proximity_sensor(self, prox_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the proximity sensor; can remove proximity sensor by passing None
        :param prox_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_proximity_sensor():
            self._delete_sensor(SensorType.PROXIMITY)
        if prox_sensor is not None:
            self._add_sensor(SensorType.PROXIMITY, prox_sensor)
        return self

    def has_relative_humidity_sensor(self) -> bool:
        """
        check if relative humidity sensor is in sensor_data_dict
        :return: True if linear relative humidity sensor exists
        """
        return SensorType.RELATIVE_HUMIDITY in self.station_data.keys()

    def has_relative_humidity_data(self) -> bool:
        """
        check if the relative humidity sensor has any data
        :return: True if relative humidity sensor has any data
        """
        return self.has_relative_humidity_sensor() and self.relative_humidity_sensor().num_samples() > 0

    def relative_humidity_sensor(self) -> Optional[SensorData]:
        """
        return the relative humidity sensor if it exists
        :return: relative humidity sensor if it exists, None otherwise
        """
        if self.has_relative_humidity_sensor():
            return self.station_data[SensorType.RELATIVE_HUMIDITY]
        return None

    def set_relative_humidity_sensor(self, relhum_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the relative humidity sensor; can remove relative humidity sensor by passing None
        :param relhum_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_relative_humidity_sensor():
            self._delete_sensor(SensorType.RELATIVE_HUMIDITY)
        if relhum_sensor is not None:
            self._add_sensor(SensorType.RELATIVE_HUMIDITY, relhum_sensor)
        return self

    def has_rotation_vector_sensor(self) -> bool:
        """
        check if rotation vector sensor is in sensor_data_dict
        :return: True if rotation vector sensor exists
        """
        return SensorType.ROTATION_VECTOR in self.station_data.keys()

    def has_rotation_vector_data(self) -> bool:
        """
        check if the rotation vector sensor has any data
        :return: True if rotation vector sensor has any data
        """
        return self.has_rotation_vector_sensor() and self.rotation_vector_sensor().num_samples() > 0

    def rotation_vector_sensor(self) -> Optional[SensorData]:
        """
        return the rotation vector sensor if it exists
        :return: rotation vector sensor if it exists, None otherwise
        """
        if self.has_rotation_vector_sensor():
            return self.station_data[SensorType.ROTATION_VECTOR]
        return None

    def set_rotation_vector_sensor(self, rotvec_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the rotation vector sensor; can remove rotation vector sensor by passing None
        :param rotvec_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_rotation_vector_sensor():
            self._delete_sensor(SensorType.ROTATION_VECTOR)
        if rotvec_sensor is not None:
            self._add_sensor(SensorType.ROTATION_VECTOR, rotvec_sensor)
        return self

    def has_compressed_audio_sensor(self) -> bool:
        """
        check if compressed audio sensor is in sensor_data_dict
        :return: True if compressed audio sensor exists
        """
        return SensorType.COMPRESSED_AUDIO in self.station_data.keys()

    def has_compressed_audio_data(self) -> bool:
        """
        check if the compressed audio sensor has any data
        :return: True if compressed audio sensor has any data
        """
        return self.has_compressed_audio_sensor() and self.compressed_audio_sensor().num_samples() > 0

    def compressed_audio_sensor(self) -> Optional[SensorData]:
        """
        return the compressed audio sensor if it exists
        :return: compressed audio sensor if it exists, None otherwise
        """
        if self.has_compressed_audio_sensor():
            return self.station_data[SensorType.COMPRESSED_AUDIO]
        return None

    def set_compressed_audio_sensor(self, compaudio_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the compressed audio sensor; can remove compressed audio sensor by passing None
        :param compaudio_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_compressed_audio_sensor():
            self._delete_sensor(SensorType.COMPRESSED_AUDIO)
        if compaudio_sensor is not None:
            self._add_sensor(SensorType.COMPRESSED_AUDIO, compaudio_sensor)
        return self

    def has_health_sensor(self) -> bool:
        """
        check if station health is in sensor_data_dict
        :return: True if station health sensor exists
        """
        return SensorType.STATION_HEALTH in self.station_data.keys()

    def has_health_data(self) -> bool:
        """
        check if the station health sensor has any data
        :return: True if station health sensor has any data
        """
        return self.has_health_sensor() and self.health_sensor().num_samples() > 0

    def health_sensor(self) -> Optional[SensorData]:
        """
        return the station health sensor if it exists
        :return: station health sensor if it exists, None otherwise
        """
        if self.has_health_sensor():
            return self.station_data[SensorType.STATION_HEALTH]
        return None

    def set_health_sensor(self, health_sensor: Optional[SensorData]) -> 'Station':
        """
        sets the station health sensor; can remove station health sensor by passing None
        :param health_sensor: the SensorData to set or None
        :return: the edited DataPacket
        """
        if self.has_health_sensor():
            self._delete_sensor(SensorType.STATION_HEALTH)
        if health_sensor is not None:
            self._add_sensor(SensorType.STATION_HEALTH, health_sensor)
        return self

    def get_non_audio_sensors(self) -> Dict[SensorType, SensorData]:
        """
        returns all non-audio sensors in the station
        :return: dict of SensorType and SensorData of all non-audio sensors in the station
        """
        return {k: v for (k, v) in self.station_data.items() if SensorType.AUDIO not in k}

    def update_station_locations(self, from_data: bool = False):
        """
        updates the station locations in the metadata
        :param from_data: if True, pulls location data from the location sensor, default False
        """
        if from_data:
            if self.has_location_sensor():
                # noinspection PyTypeChecker
                self.station_metadata.location_data.all_locations = \
                    self.location_sensor().data_df.apply(station_location_from_data, axis=1).values
            else:
                raise ValueError(f"Attempted to update locations of station {self.station_metadata.station_id} "
                                 f"using data, but there is no data!")
        else:
            self.station_metadata.location_data.all_locations = \
                [packet.best_location if packet.best_location else StationLocation() for packet in self.packet_data]

    def update_station_location_metadata(self, start_datetime: float, end_datetime: float):
        """
        update the station's location metadata using the metadata and/or location data of the station
        :param start_datetime: the start timestamp in microseconds since epoch UTC of the window to consider
        :param end_datetime: the end timestamp in microseconds since epoch UTC of the window to consider
        """
        # get locations from packet metadata
        self.update_station_locations()
        self.station_metadata.location_data.update_window_locations(start_datetime, end_datetime)
        # not enough packet metadata locations, get data locations
        if not self.station_metadata.location_data.calc_mean_and_std_from_locations() and self.has_location_data():
            self.update_station_locations(from_data=True)
            self.station_metadata.location_data.update_window_locations(start_datetime, end_datetime)
            self.station_metadata.location_data.calc_mean_and_std_from_locations()

    def packet_gap_detector(self, gap_time_s: float):
        """
        Uses the station's packet and audio data to detect gaps at least gap_time_s seconds long.
        updates the station's packet metadata if there are no gaps
        :param gap_time_s: float, minimum gap time in seconds
        """
        for packet in range(len(self.packet_data) - 1):
            data_start = self.packet_data[packet].data_start_timestamp
            data_num_samples = self.packet_data[packet].num_audio_samples
            next_packet_start_index = \
                self.audio_sensor().data_df.query("timestamps >= @data_start").first_valid_index() + data_num_samples
            data_end = self.audio_sensor().data_timestamps()[next_packet_start_index - 1]
            next_packet_start = self.audio_sensor().data_timestamps()[next_packet_start_index]
            if next_packet_start - data_end < dtu.seconds_to_microseconds(gap_time_s):
                self.packet_data[packet].micros_to_next_packet = \
                    (next_packet_start - data_start) / data_num_samples

    def update_timestamps(self):
        """
        updates the timestamps in all SensorData objects using the station_best_offset of the station_timing
        """
        if self.station_metadata.station_timing_is_corrected:
            print("WARNING: Timestamps already corrected!")
        else:
            if not self.station_metadata.timing_data:
                print("WARNING: Station does not have timing data, assuming existing values are the correct ones!")
            else:
                delta = self.station_metadata.timing_data.station_best_offset
                for sensor in self.station_data.values():
                    sensor.update_data_timestamps(delta)
                for packet in self.packet_data:
                    packet.data_start_timestamp += delta
                    packet.data_end_timestamp += delta
                    if packet.best_location:
                        packet.best_location.update_timestamps(delta)
                self.station_metadata.timing_data.station_first_data_timestamp += delta
                self.station_metadata.location_data.update_timestamps(delta)
            self.station_metadata.station_timing_is_corrected = True

    def revert_timestamps(self):
        """
        reverts the timestamps in all SensorData objects using the station_best_offset of the station_timing
        """
        if self.station_metadata.station_timing_is_corrected:
            if not self.station_metadata.timing_data:
                print("WARNING: Station does not have timing data, assuming existing values are the correct ones!")
            else:
                delta = self.station_metadata.timing_data.station_best_offset
                for sensor in self.station_data.values():
                    sensor.update_data_timestamps(-delta)
                for packet in self.packet_data:
                    packet.data_start_timestamp -= delta
                    packet.data_end_timestamp -= delta
                    if packet.best_location:
                        packet.best_location.update_timestamps(-delta)
                self.station_metadata.timing_data.station_first_data_timestamp -= delta
                self.station_metadata.location_data.update_timestamps(-delta)
                self.station_metadata.station_timing_is_corrected = False
        else:
            print("WARNING: Cannot revert timestamps that are not corrected!")



"""
Defines generic station metadata for API-independent analysis
all timestamps are integers in microseconds unless otherwise stated
"""
import numpy as np
import pandas as pd
from typing import List, Optional, Tuple, Dict
from dataclasses import dataclass, field

from redvox.common import tri_message_stats as tms
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider


@dataclass
class StationKey:
    """
    A set of values that uniquely define a station
    Properties:
        id: str, id of the station
        uuid: str, uuid of the station
        start_timestamp_micros: float, starting time of the station in microseconds since epoch UTC
    """
    id: str
    uuid: str
    start_timestamp_micros: float

    def get_key(self) -> Tuple[str, str, float]:
        return self.id, self.uuid, self.start_timestamp_micros


@dataclass
class StationLocation:
    """
    Generic StationLocation class for API-independent analysis
    Properties:
        lat_lon_timestamp: float, timestamp of the latitude and longitude, default np.nan
        altitude_timestamp: float, timestamp of the altitude, default np.nan
        speed_timestamp: float, timestamp of the speed, default np.nan
        bearing_timestamp: float, timestamp of the bearing, default np.nan
        provider: str, method/device name that provided the location, default "None"
        score: float, the value of the location's quality, default np.nan
        latitude: float, the latitude in degrees of the location, default np.nan
        longitude: float, the longitude in degrees of the location, default np.nan
        altitude: float, the altitude in meters of the location, default np.nan
        speed: float, the speed in meters/second of the location, default np.nan
        bearing: float, the bearing in degrees of the location, default np.nan
        horizontal_accuracy: float, the horizontal accuracy in meters of the location, default np.nan
        vertical_accuracy: float, the vertical accuracy in meters of the location, default np.nan
        speed_accuracy: float, the speed accuracy in meters/second of the location, default np.nan
        bearing_accuracy: float, the bearing accuracy in degrees of the location, default np.nan
    """
    lat_lon_timestamp: float = np.nan
    altitude_timestamp: float = np.nan
    speed_timestamp: float = np.nan
    bearing_timestamp: float = np.nan
    provider: str = "None"
    score: float = np.nan
    latitude: float = np.nan
    longitude: float = np.nan
    altitude: float = np.nan
    speed: float = np.nan
    bearing: float = np.nan
    horizontal_accuracy: float = np.nan
    vertical_accuracy: float = np.nan
    speed_accuracy: float = np.nan
    bearing_accuracy: float = np.nan

    def update_timestamps(self, time_delta: float):
        """
        adds the time_delta to the location's timestamps; use negative values to go backwards in time
        :param time_delta: time to add to location's timestamps
        """
        self.lat_lon_timestamp += time_delta
        self.altitude_timestamp += time_delta
        self.speed_timestamp += time_delta
        self.bearing_timestamp += time_delta

    def is_empty(self) -> bool:
        """
        :return: True if there is no lat_lon_timestamp
        """
        return np.isnan(self.lat_lon_timestamp)


def station_location_from_data(data_location: pd.Series) -> StationLocation:
    """
    function to convert a location sensor pandas series to a StationLocation
    :param data_location: a pandas series object to convert
    :return: StationLocation with data from data_location
    """
    return StationLocation(data_location["timestamps"], data_location["timestamps"],
                           data_location["timestamps"], data_location["timestamps"],
                           LocationProvider(data_location["location_provider"]).name,
                           np.nan, data_location["latitude"], data_location["longitude"],
                           data_location["altitude"], data_location["speed"], data_location["bearing"],
                           data_location["horizontal_accuracy"], data_location["vertical_accuracy"],
                           data_location["speed_accuracy"], data_location["bearing_accuracy"])


def _location_sorter(best_location: StationLocation, other_location: StationLocation, new_location: StationLocation,
                     ascending: bool = True) -> Tuple[StationLocation, StationLocation]:
    """
    Sorts up to three locations for quality purposes.
    the ascending parameter is used specifically for timestamps that come before a target time.
    :param best_location: current best location
    :param other_location: current other best location
    :param new_location: new location to compare against
    :param ascending: if True, values are sorted in ascending order, default True
    :return: the new best and other best locations
    """
    if best_location.is_empty():
        new_best_location = new_location
        new_other_location = other_location
    else:
        locs_to_sort = [best_location, other_location, new_location]
        times_to_sort = np.array([x.lat_lon_timestamp for x in locs_to_sort])
        if not ascending:
            times_to_sort *= -1
        sorted_locs = [locs_to_sort[i] for i in np.argsort(times_to_sort)]
        new_best_location = sorted_locs[0]
        new_other_location = sorted_locs[1]
    return new_best_location, new_other_location


@dataclass
class LocationData:
    """
    Location metadata statistics for the station
    Properties:
        best_location: Optional StationLocation object, the best rated location for the station, default None
        all_locations: List of StationLocation objects, all locations for the station including best, default empty list
        mean_latitude: float, the mean latitude in degrees of all locations, default np.nan
        std_latitude: float, the std dev of latitude in degrees of all locations, default 0.0
        mean_longitude: float, the mean longitude in degrees of all locations, default np.nan
        std_longitude: float, the std dev of longitude in degrees of all locations, default 0.0
        mean_altitude: float, the mean altitude in meters of all locations, default np.nan
        std_altitude: float, the std dev of altitude in meters of all locations, default 0.0
        mean_speed: float, the mean speed in meters/second of all locations, default np.nan
        std_speed: float, the std dev of speed in meters/second of all locations, default 0.0
        mean_bearing: float, the mean bearing in degrees of all locations, default np.nan
        std_bearing: float, the std dev of bearing in degrees of all locations, default 0.0
        mean_horizontal_accuracy: float, the mean horizontal accuracy in meters of all locations, default np.nan
        std_horizontal_accuracy: float, the std dev of horizontal accuracy in meters of all locations, default 0.0
        mean_vertical_accuracy: float, the mean vertical accuracy in meters of all locations, default np.nan
        std_vertical_accuracy: float, the std dev of vertical accuracy in meters of all locations, default 0.0
        mean_speed_accuracy: float, the mean speed accuracy in meters/second of all locations, default np.nan
        std_speed_accuracy: float, the std dev of speed accuracy in meters/second of all locations, default 0.0
        mean_bearing_accuracy: float, the mean bearing accuracy in degrees of all locations, default np.nan
        std_bearing_accuracy: float, the std dev of bearing accuracy in degrees of all locations, default 0.0
        mean_provider: str, method/device name that provided the mean location, default "None"
    """
    best_location: Optional[StationLocation] = None
    all_locations: List[StationLocation] = field(default_factory=list)
    mean_latitude: float = np.nan
    std_latitude: float = 0.0
    mean_longitude: float = np.nan
    std_longitude: float = 0.0
    mean_altitude: float = np.nan
    std_altitude: float = 0.0
    mean_speed: float = np.nan
    std_speed: float = 0.0
    mean_bearing: float = np.nan
    std_bearing: float = 0.0
    mean_horizontal_accuracy: float = np.nan
    std_horizontal_accuracy: float = 0.0
    mean_vertical_accuracy: float = np.nan
    std_vertical_accuracy: float = 0.0
    mean_speed_accuracy: float = np.nan
    std_speed_accuracy: float = 0.0
    mean_bearing_accuracy: float = np.nan
    std_bearing_accuracy: float = 0.0
    mean_provider: str = "None"

    def update_timestamps(self, time_delta: float):
        """
        adds the time_delta to all locations' timestamps; use negative values to go backwards in time
        :param time_delta: time to add to all locations' timestamps
        """
        for location in self.all_locations:
            location.lat_lon_timestamp += time_delta
            location.altitude_timestamp += time_delta
            location.speed_timestamp += time_delta
            location.bearing_timestamp += time_delta

    def get_sorted_all_locations(self) -> List[StationLocation]:
        """
        :return: sorted list of self.all_locations by lat_lon_timestamp
        """
        sorted_indices = np.argsort([x.lat_lon_timestamp for x in self.all_locations])
        return [self.all_locations[i] for i in sorted_indices]

    def update_window_locations(self, start_datetime: float, end_datetime: float):
        """
        Updates all_locations to be any location within the window and up to two locations outside the window.
        Locations with timestamp equal to start or end datetimes are considered outside
        :param start_datetime: the start timestamp in microseconds since epoch UTC of the window to consider
        :param end_datetime: the end timestamp in microseconds since epoch UTC of the window to consider
        """
        valid_locations: List[StationLocation] = []
        before_location: StationLocation = StationLocation()
        after_location: StationLocation = StationLocation()
        other_before_location: StationLocation = StationLocation()
        other_after_location: StationLocation = StationLocation()

        for location in self.all_locations:
            if start_datetime < location.lat_lon_timestamp < end_datetime:
                valid_locations.append(location)
            elif start_datetime >= location.lat_lon_timestamp:
                before_location, other_before_location = \
                    _location_sorter(before_location, other_before_location, location, False)
            else:
                after_location, other_after_location = _location_sorter(after_location, other_after_location, location)

        if not before_location.is_empty():
            valid_locations.append(before_location)
        if not after_location.is_empty():
            valid_locations.append(after_location)
        if len(valid_locations) < 2:
            if after_location.is_empty() and not other_before_location.is_empty():
                valid_locations.append(other_before_location)
            elif before_location.is_empty() and not other_after_location.is_empty():
                valid_locations.append(other_after_location)
        self.all_locations = valid_locations

    def calc_mean_and_std_from_locations(self, debug: bool = False) -> bool:
        """
        compute the mean and std dev from the locations in the object
        :param debug: if True, output warnings when they occur, default False
        :return: True if success, False if failed
        """
        if len(self.all_locations) > 0:
            self.mean_latitude = np.mean([x.latitude for x in self.all_locations], axis=0)
            self.std_latitude = np.std([x.latitude for x in self.all_locations], axis=0)
            self.mean_longitude = np.mean([x.longitude for x in self.all_locations], axis=0)
            self.std_longitude = np.std([x.longitude for x in self.all_locations], axis=0)
            self.mean_altitude = np.mean([x.altitude for x in self.all_locations], axis=0)
            self.std_altitude = np.std([x.altitude for x in self.all_locations], axis=0)
            self.mean_speed = np.mean([x.speed for x in self.all_locations], axis=0)
            self.std_speed = np.std([x.speed for x in self.all_locations], axis=0)
            self.mean_bearing = np.mean([x.bearing for x in self.all_locations], axis=0)
            self.std_bearing = np.std([x.bearing for x in self.all_locations], axis=0)
            self.mean_horizontal_accuracy = np.mean([x.horizontal_accuracy for x in self.all_locations], axis=0)
            self.std_horizontal_accuracy = np.std([x.horizontal_accuracy for x in self.all_locations], axis=0)
            self.mean_vertical_accuracy = np.mean([x.vertical_accuracy for x in self.all_locations], axis=0)
            self.std_vertical_accuracy = np.std([x.vertical_accuracy for x in self.all_locations], axis=0)
            self.mean_speed_accuracy = np.mean([x.speed_accuracy for x in self.all_locations], axis=0)
            self.std_speed_accuracy = np.std([x.speed_accuracy for x in self.all_locations], axis=0)
            self.mean_bearing_accuracy = np.mean([x.bearing_accuracy for x in self.all_locations], axis=0)
            self.std_bearing_accuracy = np.std([x.bearing_accuracy for x in self.all_locations], axis=0)
        else:
            if debug:
                print("WARNING: Not enough locations to process mean and std dev on!")
            return False
        return True


class TimeSyncData:
    """
    Generic TimeSync object for packets
    Properties:
        timesync_data: dataframe of the timestamps of the sync exchanges, default empty dataframe
                                Must be a multiple of 6 elements long.
        best_latency: float, best latency of the time sync exchanges, default np.nan
        best_offset: float, best offset of the time sync exchanges, default 0.0
    """
    def __init__(self, timesync_data: np.array = np.array([]),
                 best_latency: float = np.nan, best_offset: float = 0.0):
        """
        initialize the TimeSyncData object
        :param timesync_data: np.array of the timestamps of the sync exchanges, default empty np.array
                                Must be a multiple of 6 elements long.
                                May be 1 dimension, in which case will be converted into a 2-d np.ndarray
        :param best_latency: float, best latency of the time sync exchanges, default np.nan
        :param best_offset: float, best offset of the time sync exchanges, default 0.0
        """
        if len(timesync_data) > 0 and len(timesync_data.shape) != 2:
            timesync_data = np.transpose(tms.transmit_receive_timestamps_microsec(timesync_data))
        elif len(timesync_data) == 0:
            timesync_data = []
        self.timesync_data = pd.DataFrame(timesync_data, columns=["a1", "a2", "a3", "b1", "b2", "b3"])
        self.best_latency = best_latency
        self.best_offset = best_offset

    def num_samples(self) -> int:
        """
        :return: the number of time sync exchanges, which are the rows in the dataframe
        """
        return self.timesync_data.shape[0]


@dataclass
class DataPacket:
    """
    Generic DataPacket class for API-independent analysis
    Properties:
        server_received_timestamp: float, server timestamp of when data was received by the server
        app_start_timestamp: float, machine timestamp of when app started
        data_start_timestamp: float, machine timestamp of the start of the packet's data
        data_end_timestamp: float, machine timestamp of the end of the packet's data, default np.nan
        duration_s: float, duration of data packet in seconds, default 0.0
        num_audio_samples: int, number of audio samples in the data packet, default 0
        timesync: TimeSyncData object, the timesync data associated with the packet, default empty TimeSyncData object
        micros_to_next_packet: float, the length of time in microseconds to reach the next
                                packet's (in the station's data) start time, default np.nan
                                Should be close to 0 or close to self.expected_sample_interval_s
                                (when converted to same units)
        best_location: Optional StationLocation metadata, default None
    """
    server_timestamp: float
    app_start_timestamp: float
    data_start_timestamp: float
    data_end_timestamp: float = np.nan
    duration_s: float = 0.0
    num_audio_samples: int = 0
    timesync: TimeSyncData = TimeSyncData()
    micros_to_next_packet: float = np.nan
    best_location: Optional[StationLocation] = None

    def expected_sample_interval_s(self) -> float:
        """
        the packet's expected sample interval based on its own data
        :return: the packet's expected sample interval in seconds
        """
        return self.duration_s / self.num_audio_samples


@dataclass
class StationTiming:
    """
    Generic StationTiming class for API-independent analysis
    Properties:
        station_start_timestamp: float, timestamp when station started recording
        audio_sample_rate_hz: float, sample rate in hz of audio sensor
        station_first_data_timestamp: float, first timestamp chronologically of the data
        episode_start_timestamp_s: float, timestamp of start of segment of interest in seconds since epoch UTC,
                                    default np.nan
        episode_end_timestamp_s: float, timestamp of end of segment of interest in seconds since epoch UTC,
                                    default np.nan
        station_best_latency: float, best latency of data, default np.nan
        station_best_offset: float, best offset of data, default 0.0
        station_mean_offset: float, mean offset of data, default 0.0
        station_std_offset: float, std dev of offset of data, default 0.0
    """
    station_start_timestamp: float
    audio_sample_rate_hz: float
    station_first_data_timestamp: float
    episode_start_timestamp_s: float = np.nan
    episode_end_timestamp_s: float = np.nan
    station_best_latency: float = np.nan
    station_best_offset: float = 0.0
    station_mean_offset: float = 0.0
    station_std_offset: float = 0.0


@dataclass
class StationMetadata:
    """
    Generic StationMetadata class for API-independent analysis
    Properties:
        station_id: str, id of the station
        station_make: str, maker of the station
        station_model: str, model of the station
        station_timing_is_corrected: bool, if True, the station's timestamps have been altered from their raw values
                                        default False
        station_os: optional str, operating system of the station, default None
        station_os_version: optional str, station OS version, default None
        station_app: optional str, the name of the recording software used by the station, default None
        station_app_version: optional str, the recording software version, default None
        is_mic_scrambled: optional bool, True if mic data is scrambled, default False
        timing_data: optional StationTiming metadata, default None
        station_calib: optional float, station calibration value, default None
        station_network_name: optional str, name/code of network station belongs to, default None
        station_name: optional str, name/code of station, default None
        station_location_name: optional str, name/code of location station is at, default None
        station_channel_name: optional str, name/code of channel station is recording, default None
        station_channel_encoding: optional str, name/code of channel encoding method, default None
        station_uuid: optional str, uuid of the station, default is the same value as station_id
        best_location: optional StationLocation metadata, default empty LocationData
    """
    station_id: str
    station_make: str
    station_model: str
    station_timing_is_corrected: bool = False
    station_os: Optional[str] = None
    station_os_version: Optional[str] = None
    station_app: Optional[str] = None
    station_app_version: Optional[str] = None
    is_mic_scrambled: Optional[bool] = False
    timing_data: Optional[StationTiming] = None
    station_calib: Optional[float] = None
    station_network_name: Optional[str] = None
    station_name: Optional[str] = None
    station_location_name: Optional[str] = None
    station_channel_name: Optional[str] = None
    station_channel_encoding: Optional[str] = None
    station_uuid: Optional[str] = None
    location_data: Optional[LocationData] = None

    def __post_init__(self):
        """
        if the station_uuid is None, set it to be station_id
        if the location data is None, set it to empty LocationData
        """
        if not self.station_uuid:
            self.station_uuid = self.station_id
        if not self.location_data:
            self.location_data = LocationData()
